from .core.get_filter import get_filter
from .core.t2w import t2w
from .fourier_coeff import load_fourier_coeff
from gwpy.timeseries import TimeSeries
from wdm_wavelet.types import TimeFrequencyMap


class WDM:
    """
    WDM class
    """

    def __init__(self, M, K, beta_order, precision):
        self.M = M
        self.K = K
        self.beta_order = beta_order
        self.precision = precision

        Cos, Cos2, SinCos, CosSize, Cos2Size, SinCosSize = load_fourier_coeff()

        n_max = int(3e5)
        filter = get_filter(M, K, beta_order, n_max, Cos, CosSize)

        residual = 1 - filter[0] ** 2
        prec = 10 ** (-precision)

        N = 1
        M2 = M * 2
        while residual > prec or (N - 1) % M2 or N // M2 < 3:
            residual -= 2 * filter[N] ** 2
            N += 1

        self.filter = filter[:N]
        self.m_H = N

    def t2w(self, strain, sample_rate=None, t0=0, MM=-1):
        """
        Transform time series to WDM


        Parameters
        ----------
        strain: array-like, TimeSeries
            Time series data

        sample_rate: float, optional
            Sample rate of the time series data. Required if strain is not a TimeSeries object

        t0: float, optional
            Start time of the time series data. Default is 0

        MM: int, optional
            MM = 0 requests power map of combined quadratures (not amplitudes for both)
        """

        # convert to plain numpy array for numba compatibility
        if isinstance(strain, TimeSeries):
            sample_rate = strain.sample_rate.value
            t0 = strain.t0.value
            strain = strain.value
        elif str(type(strain)) == "<class 'pycbc.types.timeseries.TimeSeries'>":
            try:
                sample_rate = strain.sample_rate
                t0 = int(strain.start_time)
                strain = strain.data
            except AttributeError:
                raise ValueError("strain is detected as a pycbc TimeSeries object, "
                                 "but the sample_rate or start_time cannot be extracted")
        else:
            if sample_rate is None:
                raise ValueError("sample_rate must be provided if strain is not"
                                 " a gwpy.TimeSeries object or a pycbc.TimeSeries object")

        # calculate the time frequency map
        _, _, time_frequency_map = t2w(self.M, self.m_H, strain, self.filter, MM)

        # calculate the df and dt
        max_f_layer = time_frequency_map.shape[2] - 1
        max_t_layer = time_frequency_map.shape[1]
        df = sample_rate/max_f_layer/2.

        t_len = len(strain) / sample_rate
        dt = t_len/max_t_layer

        time_frequency_map = time_frequency_map[0] - 1j * time_frequency_map[1]

        return TimeFrequencyMap(time_frequency_map.T, df, dt, t0)

    def w2t(self):
        raise NotImplementedError("w2t is not implemented yet")


import numpy as np


def test_wdm():
    # Create a WDM object
    from wdm_wavelet.wdm import WDM

    fake_strain = np.random.randn(4096)

    wdm = WDM(32, 64, 6, 10)
    tf_map = wdm.t2w(fake_strain, sample_rate=4096.)

    assert len(tf_map.data) == tf_map.n_freq
    assert tf_map.df == 4096./32/2.



import numpy as np
from numpy.fft import fft
from numba import njit


@njit(cache=True)
def t2w(M, m_H, WWS, filter, MM):
    """
    Transform time series to WDM

    :param M: max layers
    :param m_H:
    :param WWS: time series
    :param nWWS: length of time series
    :param filter: filter coefficients
    :param MM:  MM = 0 requests power map of combined quadratures (not amplitudes for both)
    :return:
    """

    M1 = M + 1
    M2 = M * 2
    nWDM = m_H
    nTS = len(WWS)
    KK = MM

    if MM <= 0:
        MM = M

    # adjust nWWS to be a multiple of MM
    nWWS = len(WWS)
    # this->nWWS += this->nWWS%MM ? MM-this->nWWS%MM : 0;
    nWWS += MM - nWWS % MM if nWWS % MM else 0

    # initialize time series with boundary conditions (mirror)
    m = nWWS + 2 * nWDM
    ts = np.zeros(m)

    for n in range(nWDM):
        ts[nWDM - n] = WWS[n]
    for n in range(nTS):
        ts[nWDM + n] = WWS[n]
    for n in range(int(m - nWDM - nTS)):
        ts[n + nWDM + nTS] = WWS[nTS - n - 1]

    # create symmetric arrays
    wdm = filter[:nWDM]
    # INV = np.array(filter[nWDM - 1::-1])

    # WDM = INV[::-1]

    # reallocate TF array
    N = int(nWWS / MM)
    L = 2 * N * M1 if KK < 0 else N * M1
    m_L = m_H if KK < 0 else 0

    pWDM = np.zeros(L)  # Assuming pWWS is a numpy array

    odd = 0
    sqrt2 = np.sqrt(2)

    for n in range(N):
        # create references
        map00 = pWDM[n * M1:]
        map90 = pWDM[(N + n) * M1:]
        pTS = ts[nWDM + n * MM:]

        re = np.zeros(M2)
        im = np.zeros(M2)

        J = M2
        for j in range(0, nWDM - 1, M2):
            J = M2 + j
            pTS_inv = ts[nWDM + n * MM - J:]
            for m in range(M2):
                re[m] += pTS[j + m] * wdm[j + m] + pTS_inv[m] * wdm[J - m]

        re[0] += wdm[J] * pTS[J]

        # Perform FFT
        fft_result = fft(re)
        re, im = fft_result.real, fft_result.imag

        re[0] = im[0] = re[0] / sqrt2
        re[M] = im[M] = re[M] / sqrt2

        if KK < 0:
            for m in range(M + 1):
                if (m + odd) & 1:
                    map00[m] = sqrt2 * im[m]
                    map90[m] = sqrt2 * re[m]
                else:
                    map00[m] = sqrt2 * re[m]
                    map90[m] = -sqrt2 * im[m]
        else:  # power map
            map00[:M + 1] = re[:M + 1] ** 2 + im[:M + 1] ** 2

        odd = 1 - odd

    return m_L, nWWS, pWDM.reshape(2, N, M1)



from dataclasses import dataclass
import numpy as np


@dataclass
class TimeFrequencyMap:
    """
    Time-frequency map data class

    Parameters
    ----------
    data: np.ndarray
        Complex Time-frequency map data
    df: float
        Frequency resolution
    dt: float
        Time resolution
    t0: float
        Start time
    """
    data: np.ndarray
    df: float
    dt: float
    t0: float

    @property
    def n_freq(self):
        """
        Number of frequency bins

        Returns
        -------
        int
        """
        return self.data.shape[0]

    @property
    def n_time(self):
        """
        Number of time bins

        :returns: int
        """
        return self.data.shape[1]

    @property
    def start_time(self):
        """
        Start time

        :return: float
        """
        return self.t0

    @property
    def end_time(self):
        """
        End time

        :return: float
        """
        return self.t0 + self.n_time * self.dt

    @property
    def start_freq(self):
        """
        Start frequency

        :return: float
        """
        return 0.

    @property
    def end_freq(self):
        """
        End frequency

        :return: float
        """
        return self.df * self.n_freq

    def energy(self, low_cut=None):
        """
        Energy of the time-frequency map

        :param low_cut: float, optional
            Low cut frequency. If provided, the energy below this frequency will be set to NaN
        :return: np.ndarray
        """
        energy = np.abs(self.data)
        if low_cut is not None:
            energy = np.where(energy < low_cut, np.nan, energy)
        return energy

    def plot_energy(self, fig=None, ax=None, low_cut=None, **kwargs):
        """
        Plot the energy of the time-frequency map

        :param fig: matplotlib.figure.Figure, optional
        :param ax: matplotlib.axes.Axes, optional
        :param low_cut: float, optional
            Low cut frequency. If provided, the energy below this frequency will be set to NaN and not plotted
        :param kwargs: dict
            Additional arguments to pass to ax.imshow
        :return: matplotlib.axes.Axes
        """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots()
        im = ax.imshow(self.energy(low_cut=low_cut), aspect="auto", origin="lower",
                  extent=[self.start_time, self.end_time, self.start_freq, self.end_freq],
                  **kwargs)

        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Frequency (Hz)")

        fig.colorbar(im, ax=ax)
        return ax


import math
import numpy as np
from numba import njit


@njit(cache=True)
def get_filter(M, K, BetaOrder, n, Cos, CosSize):
    # Extract necessary attributes from the wdm
    B = math.pi / K
    A = (K - M) * math.pi / (2.0 * K * M)
    K2 = K ** 2
    gNorm = math.sqrt(2 * M) / math.pi

    # Initialize the filter array
    filter = np.zeros(n)

    # Get the Fourier array
    Fourier = Cos[BetaOrder]
    nFourier = CosSize[BetaOrder]
    fourier = np.copy(Fourier)

    # Calculate filter coefficients
    filter[0] = (A + fourier[0] * B) * gNorm
    fNorm = filter[0] ** 2
    fourier[0] /= math.sqrt(2)

    for i in range(1, n):
        di = float(i)
        i2 = di * di
        sumEven = 0
        sumOdd = 0

        # Calculating the sums for even and odd
        for j in range(nFourier):
            if i % K:
                pass
            elif j == i // K:
                continue
            coeff = di / (i2 - j ** 2 * K2)
            if j & 1:
                sumOdd += coeff * fourier[j]
            else:
                sumEven += coeff * fourier[j]

        # Calculating intAB
        intAB = 0
        if i % K == 0 and i // K < nFourier:
            intAB = fourier[i // K] * B / 2 * math.cos(i * A)

        intAB += 2 * (sumEven * math.sin(i * B / 2) * math.cos(i * math.pi / (2 * M)) -
                      sumOdd * math.sin(i * math.pi / (2 * M)) * math.cos(i * B / 2))

        filter[i] = gNorm * (math.sin(i * A) / di + math.sqrt(2) * intAB)
        fNorm += 2 * filter[i] ** 2

    return filter



import os
from setuptools import setup
from setuptools import find_packages

requires = []
install_requires = [
    "numpy",
    "numba",
    "rocket-fft",
    "gwpy",
]


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def find_files(dirname, relpath=None):
    def find_paths(dirname):
        items = []
        for fname in os.listdir(dirname):
            path = os.path.join(dirname, fname)
            if os.path.isdir(path):
                items += find_paths(path)
            elif not path.endswith(".py") and not path.endswith(".pyc"):
                items.append(path)
        return items

    items = find_paths(dirname)
    if relpath is None:
        relpath = dirname
    print(items)
    return [os.path.relpath(path, relpath) for path in items]


extensions = [
]


setup(
    name="cWB Burst Waveform",
    author="Yumeng Xu",
    author_email="yumeng.xu@ligo.org",
    description=("python waveform model for cWB burst waveforms"),
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    keywords=['ligo', 'physics', 'gravity', 'signal processing', 'gravitational waves'],
    url="https://git.ligo.org/yumeng.xu/setup.py",
    install_requires=install_requires,
    packages=find_packages(),
    include_package_data=True,
    # ext_modules=cythonize(extensions),
    python_requires='>=3.8'
)